#!/usr/bin/env python3
import argparse
from pathlib import Path
from mpk import FILE_EXT


def run():
    parser = argparse.ArgumentParser(description="MPK_EXTRACT")
    parser.add_argument(
        "input", type=Path, nargs="*", default=".", help="input file or directory"
    )
    parser.add_argument(
        "-t", help="input type of audio", default="track", choices=["track", "sample"]
    )
    parser.add_argument("--ext", help="extensions to include", choices=FILE_EXT)
    parser.add_argument(
        "--json", type=Path, help="write output to json", default="result.json"
    )
    parser.add_argument(
        "--db",
        action=argparse.BooleanOptionalAction,
        help="write output to MDB",
        default=True,
    )
    parser.add_argument(
        "-d",
        default="all",
        help="descriptors to include",
        choices=[
            "lowlevel",
            "rhythm",
            "sfx",
            "tonal",
            "spectograms",
            "metadata",
            "all",
        ],
    )
    parser.add_argument("--version", action="version", version="%(prog)s 0.1.0")

    args = parser.parse_args()


if __name__ == "__main__":
    run()



try:
    from ._mpk import lib, ffi
except ImportError:
    print("mpk python bindings missing")

NULL = ffi.NULL


class Config:
    def __init__(self):
        self.fs = lib.mpk_fs_config_new(NULL)
        self.db = lib.mpk_db_config_new()
        self.jack = lib.mpk_jack_config_new()
        self.cfg = lib.mpk_config_new(self.fs, self.db, self.jack)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        lib.mpk_fs_config_free(self.fs)
        lib.mpk_db_config_free(self.db)
        lib.mpk_jack_config_free(self.jack)

    def write(self, file):
        print("writing config to " + file)
        lib.mpk_config_write(self.cfg, file.encode())

    def load(self, file):
        print("loaded config from " + file)
        self.cfg = lib.mpk_config_load(file.encode())

    def db_flags(self):
        return lib.mpk_db_config_flags(self.db)

    def get_path(self, path):
        return ffi.string(lib.mpk_fs_config_get_path(self.fs, path.encode())).decode()


class Mdb:
    def __init__(self):
        self.db = lib.mdb_new(NULL)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        lib.mdb_free(self.db)

    def init(self):
        lib.mdb_init(self.db)
        print("initialized Mdb")

    def exec_batch(self, sql):
        return lib.mdb_exec_batch(self.db, sql.encode())

    def insert_track(self, track):
        return lib.mdb_insert_track(self.db, track.encode())

    def insert_track_tags(self, id, tags, ptr=True):
        print("inserting tags:", tags, "for track_id:", id)
        return lib.mdb_insert_track_tags(self.db, id, tags)

    def insert_track_tags_musicbrainz(self, id, tags, ptr=True):
        print("inserting musicbrainz_tags:", tags, "for track_id:", id)
        return lib.mdb_insert_track_tags_musicbrainz(self.db, id, tags)

    def insert_track_featues_lowlevel(self, id, features, ptr=True):
        print("inserting lowlevel_features for track_id:", id)
        return lib.mdb_insert_track_features_lowlevel(self.db, id, features)

    def insert_track_features_rhythm(self, id, features, ptr=True):
        print("inserting rhythm_features for track_id:", id)
        return lib.mdb_insert_track_features_rhythm(self.db, id, features)

    def insert_track_features_sfx(self, id, features, ptr=True):
        print("inserting sfx_features for track_id:", id)
        return lib.mdb_insert_track_features_sfx(self.db, id, features)

    def insert_track_features_tonal(self, id, features, ptr=True):
        print("inserting tags for track_id:", id)
        return lib.mdb_insert_track_features_tonal(self.db, id, features)

    def insert_track_images(self, id, images, ptr=True):
        print("inserting spectograms for track_id:", id)
        return lib.mdb_insert_track_images(self.db, id, images)


def vectorize(arr):
    buf = ffi.from_buffer("float[]", arr)
    return lib.mdb_vecreal_new(ffi.cast("const float *", buf), len(buf))


def track_tags(tags):
    tags[0:4] = [x.encode() for x in tags[0:4]]
    return lib.mdb_track_tags_new(*tags)


def musicbrainz_tags(tags):
    tags = [x.encode() for x in tags]
    return lib.mdb_musicbrainz_tags_new(*tags)


def lowlevel_features(features):
    features = [vectorize(x) for x in features[1:]]
    return lib.mdb_lowlevel_features_new(*features)


def rhythm_features(features):
    features[3] = vectorize(features[3])
    features[10:] = [vectorize(x) for x in features[10:]]
    return lib.mdb_rhythm_features_new(*features)


def sfx_features(features):
    features[4:] = [vectorize(x) for x in features[4:]]
    return lib.mdb_sfx_features_new(*features)


def tonal_features(features):
    features[7:11] = [vectorize(x) for x in features[7:11]]
    features[11:16] = [x.encode() for x in features[11:16]]
    return lib.mdb_tonal_features_new(*features)


def spectograms(specs):
    specs = [vectorize(x) for x in specs]
    return lib.mdb_spectograms_new(*specs)



from .lib import (
    Config,
    Mdb,
    vectorize,
    track_tags,
    musicbrainz_tags,
    lowlevel_features,
    rhythm_features,
    sfx_features,
    tonal_features,
    spectograms,
)
from .extract import AudioFile, Extract, FILE_EXT



import essentia
import essentia.standard as es
from librosa import load, to_mono
import numpy as np
import os
import fnmatch

FILE_EXT = (".mp3", ".flac", ".ogg", ".wav", ".aiff", ".alac", ".aac", ".mp4")


def walk_dir(dir, exts=FILE_EXT):
    for root, dirs, files in os.walk(os.path.expanduser(dir)):
        for base in files:
            if base.lower().endswith(exts):
                yield os.path.join(root, base)


def collect_files(dir, exts=FILE_EXT):
    files = [f for f in walk_dir(dir, exts)]
    print("found %d audio files in %s" % (len(files), dir))
    return files


def isMatch(name, patterns):
    if not patterns:
        return False
    for pattern in patterns:
        if fnmatch.fnmatch(name, pattern):
            return True
    return False


def add_to_dict(dict, keys, value):
    for key in keys[:-1]:
        dict = dict.setdefault(key, {})
    dict[keys[-1]] = value


def pool_to_dict(pool, include_descs=None, ignore_descs=None):
    """a workaround to convert Pool to dict"""
    descs = pool.descriptorNames()
    if include_descs:
        descs = [d for d in descs if isMatch(d, include_descs)]
    if ignore_descs:
        descs = [d for d in descs if not isMatch(d, ignore_descs)]
    result = {}
    for d in descs:
        keys = d.split(".")
        value = pool[d]
        if type(value) is np.ndarray:
            value = value.tolist()
        add_to_dict(result, keys, value)
    return result


def bulk_extract(files, sr=44100, mono=False):
    result = {}
    for f in files:
        try:
            with Extract(f, sr, mono) as extractor:
                print("extracting:", f)
                extractor.metadata()
                extractor.features()
                extractor.mel_spec()
                extractor.freq_spec()
                extractor.log_spec()
                extractor.inverse_spec()
                result.update({f: pool_to_dict(extractor.pool)})
        except Exception as e:
            print(str(e))
    return result


class AudioFile(object):
    def __init__(self, file, sr=44100, mono=False):
        audio, sr = load(file, sr, mono)
        self.path = os.path.expanduser(file)
        self.sr = sr
        self.audio = audio

    def mono(self):
        return to_mono(self.audio)


class Extract(AudioFile):
    def __init__(self, file, sr=44100, mono=False):
        super().__init__(file, sr, mono)
        self.pool = essentia.Pool()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            print(f"exc_type: {exc_type}")
            print(f"exc_value: {exc_value}")
            print(f"exc_traceback: {exc_traceback}")

    def metadata(self):
        self.pool.merge(es.MetadataReader(filename=self.path)()[7])
        print("added metadata to pool.")

    def features(self):
        extractor = es.Extractor()
        self.pool.merge(extractor(self.mono()))
        print("added features to pool.")

    def freq_spec(self):
        windowing = es.Windowing(type="blackmanharris62", zeroPadding=2048)
        spectrum = es.Spectrum()
        amp2db = es.UnaryOperator(type="lin2db", scale=2)
        for frame in es.FrameGenerator(self.mono(), frameSize=2048, hopSize=1024):
            frame_spec = spectrum(windowing(frame))
        self.pool.add("freq_spec", amp2db(frame_spec))
        print("added freq_spec to pool.")

    def mel_spec(self):
        windowing = es.Windowing(type="blackmanharris62", zeroPadding=2048)
        spectrum = es.Spectrum()
        melbands = es.MelBands(
            numberBands=96, lowFrequencyBound=0, highFrequencyBound=11000
        )
        amp2db = es.UnaryOperator(type="lin2db", scale=2)
        for frame in es.FrameGenerator(self.mono(), frameSize=2048, hopSize=1024):
            frame_spec = spectrum(windowing(frame))
            frame_mel = melbands(frame_spec)
        self.pool.add("mel_spec", amp2db(frame_mel))
        print("added mel_spec to pool.")

    def log_spec(self):
        windowing = es.Windowing(type="blackmanharris62", zeroPadding=2048)
        spectrum = es.Spectrum()
        logfreq = es.LogSpectrum(binsPerSemitone=1)
        amp2db = es.UnaryOperator(type="lin2db", scale=2)
        for frame in es.FrameGenerator(self.mono(), frameSize=2048, hopSize=1024):
            frame_spec = spectrum(windowing(frame))
            frame_logfreq, _, _ = logfreq(frame_spec)
        self.pool.add("log_spec", amp2db(frame_logfreq))
        print("added log_spec to pool.")

    def inverse_spec(self):
        frameSize = 1024
        hopSize = 512
        fs = 44100
        w = es.Windowing(type="hamming", normalized=False)
        # make sure these are same for MFCC and IDCT computation
        NUM_BANDS = 26
        DCT_TYPE = 2
        LIFTERING = 0
        NUM_MFCCs = 13
        spectrum = es.Spectrum()
        mfcc = es.MFCC(
            numberBands=NUM_BANDS,
            numberCoefficients=NUM_MFCCs,  # make sure you specify first N mfcc: the less, the more lossy (blurry) the smoothed mel spectrum will be
            weighting="linear",  # computation of filter weights done in Hz domain (optional)
            normalize="unit_max",  #  htk filter normaliation to have constant height = 1 (optional)
            dctType=DCT_TYPE,
            logType="log",
            liftering=LIFTERING,
        )  # corresponds to htk default CEPLIFTER = 22
        idct = es.IDCT(
            inputSize=NUM_MFCCs,
            outputSize=NUM_BANDS,
            dctType=DCT_TYPE,
            liftering=LIFTERING,
        )
        all_melbands_smoothed = []
        for frame in es.FrameGenerator(
            self.mono(), frameSize=frameSize, hopSize=hopSize
        ):
            spect = spectrum(w(frame))
            melbands, mfcc_coeffs = mfcc(spect)
            melbands_smoothed = np.exp(
                idct(mfcc_coeffs)
            )  # inverse the log taken in MFCC computation
            all_melbands_smoothed.append(melbands_smoothed)
        # transpose to have it in a better shape
        # we need to convert the list to an essentia.array first (== numpy.array of floats)
        # mfccs = essentia.array(pool['MFCC']).T

        self.pool.add("inverse_spec", essentia.array(all_melbands_smoothed).T)
        print("added inverse_spec to pool.")

    def rhythm(self):
        rhythm_extractor = es.RhythmExtractor2013(method="multifeature")
        bpm, beats, confidence, _, intervals = rhythm_extractor(self.mono())
        self.pool.add("bpm", bpm[0])
        self.pool.add("beats", beats)
        self.pool.add("bpm_confidence", confidence)
        self.pool.add("bpm_intervals", intervals)
        print("added rhythm descriptors to pool.")

    def loudness(self):
        windowing = es.Windowing(type="blackmanharris62", zeroPadding=2048)
        spectrum = es.Spectrum()
        rms = es.RMS()
        hfc = es.HFC()
        for frame in es.FrameGenerator(self.mono(), frameSize=2048, hopSize=1024):
            frame_spectrum = spectrum(windowing(frame))
            self.pool.add("rms", rms(frame))
            self.pool.add("rms_spectrum", rms(frame_spectrum))
            self.pool.add("hfc", hfc(frame_spectrum))
        print("added loudness descriptors to pool.")

    def write_json(self, out_file, format="json"):
        print("writing to file: ", out_file)
        es.YamlOutput(filename=out_file, format=format, writeVersion=False)(self.pool)

    def descriptors(self):
        return self.pool.descriptorNames()

    def get(self, key):
        return self.pool.__getitem__(key)

    def remove(self, key):
        self.pool.remove(key)
        print("removed '" + key + "' from pool")

    def clear(self):
        self.pool.clear()
        print("cleared the pool.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mpk']

package_data = \
{'': ['*']}

install_requires = \
['essentia==2.1b6', 'librosa>=0.9.1,<0.10.0', 'numpy>=1.21.5,<1.22']

entry_points = \
{'console_scripts': ['main = main:run']}

setup_kwargs = {
    'name': 'mpk',
    'version': '0.1.0a6',
    'description': 'Media Programming Kit Python bindings',
    'long_description': None,
    'author': 'ellis',
    'author_email': 'ellis@rwest.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://hg.rwest.io/mpk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)
